import os
import zipfile
from typing import Dict, List

from ibm_watson_machine_learning.client import APIClient
from mlflow.exceptions import ENDPOINT_NOT_FOUND, MlflowException
from tabulate import tabulate


def list_models(client: APIClient) -> List[Dict]:
    """lists models in WML repository

    Parameters
    ----------
    client : APIClient
        WML client

    Returns
    -------
    List[Dict]
        list of model details dictionary
    """
    models = client.repository.get_model_details(get_all=True)["resources"]

    for model in models:
        model["name"] = model["metadata"]["name"]

    return models


def list_deployments(client: APIClient) -> List[Dict]:
    """lists WML deployments

    Parameters
    ----------
    client : APIClient
        WML client

    Returns
    -------
    List[Dict]
        list of deployment details dictionary
    """
    deployments = client.deployments.get_details(get_all=True)["resources"]

    # `name` is a required key in each deployment
    for deployment in deployments:
        deployment["name"] = deployment["entity"]["name"]

    return deployments


def get_deployment(client: APIClient, name: str) -> Dict:
    """retreive deployment details

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment

    Returns
    -------
    Dict
        deployment details dictionary
    """
    deployments = list_deployments(client=client)

    try:
        return next(item for item in deployments if item["entity"]["name"] == name)

    except StopIteration as _:
        raise MlflowException(
            message=f"no deployment by the name {name} exists",
            error_code=ENDPOINT_NOT_FOUND,
        )


def get_deployment_id_from_deployment_name(
    client: APIClient, deployment_name: str
) -> str:
    """Returns deployment ID from deployment name

    Parameters
    ----------
    deployment_name : str
        deployment name

    Returns
    -------
    str
        deployment id
    """
    return get_deployment(client=client, name=deployment_name)["metadata"]["id"]


def get_model_id_from_model_name(client: APIClient, model_name: str) -> str:
    """Returns model ID from model name

    Parameters
    ----------
    model_name : str
        model name

    Returns
    -------
    str
        model id
    """
    models = list_models(client=client)

    try:
        return next(item for item in models if item["name"] == model_name)["metadata"][
            "id"
        ]
    except StopIteration as _:
        raise MlflowException(
            message=f"model {model_name} not found", error_code=ENDPOINT_NOT_FOUND
        )


def get_space_id_from_space_name(client: APIClient, space_name: str) -> str:
    """Returns space ID from the space name

    Parameters
    ----------
    client : APIClient
        WML client
    space_name : str
        space name

    Returns
    -------
    str
        space id
    """
    spaces = client.spaces.get_details()["resources"]

    try:
        return next(item for item in spaces if item["entity"]["name"] == space_name)[
            "metadata"
        ]["id"]
    except StopIteration as _:
        raise MlflowException(
            message=f"space {space_name} not found", error_code=ENDPOINT_NOT_FOUND
        )


def get_model_id_from_model_details(client: APIClient, model_details: Dict) -> str:
    """Return model ID from model details dictionary

    Parameters
    ----------
    model_details : Dict
        model details dictionary

    Returns
    -------
    str
        model id
    """
    model_id = client.repository.get_model_id(model_details=model_details)
    return model_id


def get_deployment_id_from_deployment_details(
    client: APIClient, deployment_details: Dict
) -> str:
    """Return deployment ID from deployment details dictionary

    Parameters
    ----------
    client : APIClient
        WML client
    deployment_details : Dict
        deployment details dictionary

    Returns
    -------
    str
        deployment id
    """
    deployment_id = client.deployments.get_id(deployment_details=deployment_details)
    return deployment_id


def deployment_exists(client: APIClient, name: str) -> bool:
    """Checks if a deployment by the given name exists

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment

    Returns
    -------
    bool
        True if the deployment exists else False
    """
    deployments = list_deployments(client=client)
    return any(item for item in deployments if item["name"] == name)


def model_exists(client: APIClient, name: str) -> bool:
    """Checks if a model by the given name exists

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the model

    Returns
    -------
    bool
        True if the model exists else False
    """
    models = list_models(client=client)

    return any(item for item in models if item["name"] == name)


def print_package_specifications(
    client: APIClient, software_spec: str = "runtime-22.2-py3.10"
) -> None:
    """Print the package spec along with version to use as reference for creating an env

    Parameters
    ----------
    client : APIClient
        WML Client
    software_spec : str, optional
        Software Spec, by default "runtime-22.2-py3.10"
    """
    pkg_specs = client.software_specifications.get_details(
        client.software_specifications.get_id_by_name(software_spec)
    )["entity"]["software_specification"]["software_configuration"]["included_packages"]

    print(tabulate(pkg_specs, headers="keys"))


def list_endpoints(client: APIClient) -> List[Dict]:
    endpoints = client.spaces.get_details(get_all=True)["resources"]
    return endpoints


def get_endpoint(client: APIClient, endpoint: str):
    deployment_space_id = get_space_id_from_space_name(
        client=client, space_name=endpoint
    )
    endpoint_details = client.spaces.get_details(space_id=deployment_space_id)
    return endpoint_details


def list_software_specs(client: APIClient) -> List[Dict]:
    sw_specs = client.software_specifications.get_details()["resources"]
    return sw_specs


def get_software_spec(client: APIClient, sw_spec: str) -> Dict:
    sw_specs = list_software_specs(client=client)

    try:
        return next(item for item in sw_specs if item["metadata"]["name"] == sw_spec)[
            "metadata"
        ]["id"]
    except StopIteration as _:
        raise MlflowException(
            message=f"Software Specifiction - {sw_spec} not found",
            error_code=ENDPOINT_NOT_FOUND,
        )


def software_spec_exists(client: APIClient, sw_spec: str) -> bool:
    sw_specs = list_software_specs(client=client)

    return any(item for item in sw_specs if item["metadata"]["name"] == sw_spec)


def delete_sw_spec(client, name):
    sw_spec_id = client.software_specifications.get_id_by_name(name)
    client.software_specifications.delete(sw_spec_id)


def is_zipfile(file_path: str) -> bool:
    if not os.path.isfile(file_path):
        return False

    try:
        with zipfile.ZipFile(file_path) as zf:
            return zf.testzip() is None
    except zipfile.BadZipFile:
        return False



import logging
from typing import Any, Dict

from ibm_watson_machine_learning.client import APIClient
from mlflow.exceptions import MlflowException

from mlflow_watsonml.utils import *

LOGGER = logging.getLogger(__name__)


def store_model(
    client: APIClient,
    model_object: Any,
    software_spec_uid: str,
    name: str,
    model_description: str,
    model_type: str,
) -> Dict:
    """Store model_object in a WML repository

    Parameters
    ----------
    client : APIClient
        WML client
    model_object : Any
        artifact object
    software_spec_uid : str
        uid of software specification
    name : str
        name of the deployment
    model_description : str
        model description
    model_type : str
        type of model

    Returns
    -------
    Dict
        model details dictionary
    """
    model_props = {
        client.repository.ModelMetaNames.NAME: name,
        client.repository.ModelMetaNames.DESCRIPTION: model_description,
        client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: software_spec_uid,
        client.repository.ModelMetaNames.TYPE: model_type,
    }

    try:
        model_details = client.repository.store_model(
            model=model_object,
            meta_props=model_props,
            training_data=None,
            training_target=None,
            feature_names=None,
            label_column_names=None,
        )
        LOGGER.info(model_details)
        LOGGER.info(f"Stored model {name} in the repository.")

    except Exception as e:
        raise MlflowException(e)

    return model_details


def deploy_model(
    client: APIClient, name: str, model_id: str, batch: bool = False
) -> Dict:
    """Create a new WML deployment

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the new deployment to create
    model_id : str
        UID of the model stored in WML repository
    batch : bool, optional
        whether to use batch or online method of deployment,
        by default False

    Returns
    -------
    Dict
        deployment details dictionary
    """

    if batch:
        deployment_props = {
            client.deployments.ConfigurationMetaNames.NAME: name,
            client.deployments.ConfigurationMetaNames.BATCH: {},
            client.deployments.ConfigurationMetaNames.HARDWARE_SPEC: {},
        }
    else:
        deployment_props = {
            client.deployments.ConfigurationMetaNames.NAME: name,
            client.deployments.ConfigurationMetaNames.ONLINE: {},
        }

    try:
        deployment_details = client.deployments.create(
            artifact_uid=model_id,
            meta_props=deployment_props,
        )

        deployment_details["name"] = deployment_details["entity"]["name"]

        LOGGER.info(deployment_details)
        LOGGER.info(f"Created {'batch' if batch else 'online'} deployment {name}")

    except Exception as e:
        raise MlflowException(e)

    return deployment_details


def delete_deployment(client: APIClient, name: str):
    """Delete an existing deployment from WML

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment to delete
    """
    try:
        deployment_id = get_deployment_id_from_deployment_name(
            client=client, deployment_name=name
        )
        client.deployments.delete(deployment_uid=deployment_id)

        LOGGER.info(f"Deleted deployment {name} with id {deployment_id}.")
    except Exception as e:
        raise MlflowException(e)


def delete_model(client: APIClient, name: str):
    """Delete a model from WML repository

    Parameters
    ----------
    client : APIClient
        WML client
    name : str
        name of the deployment to delete
    """
    try:
        model_id = get_model_id_from_model_name(client=client, model_name=name)
        client.repository.delete(artifact_uid=model_id)

        LOGGER.info(f"Deleted model {name} with id {model_id} from the repository.")

    except Exception as e:
        raise MlflowException(e)


def update_model(client: APIClient):
    raise NotImplementedError()


def set_deployment_space(client: APIClient, deployment_space_name: str) -> APIClient:
    try:
        space_uid = get_space_id_from_space_name(
            client=client,
            space_name=deployment_space_name,
        )
        client.set.default_space(space_uid=space_uid)

        LOGGER.info(f"Set deployment space to {deployment_space_name}")

    except Exception as e:
        raise MlflowException(
            f"Failed to set deployment space {deployment_space_name}", f"{e}"
        )

    return client


def create_custom_software_spec(
    client: APIClient,
    name: str,
    base_sofware_spec: str,
    custom_packages: List[Dict[str, str]],
    rewrite: bool = False,
) -> str:
    if software_spec_exists(client=client, sw_spec=name):
        if rewrite:
            delete_sw_spec(client=client, name=name)
        else:
            LOGGER.warn(
                f"""Software spec {name} already exists."""
                """skipping software spec creation."""
                """restart with rewrite=True if software spec needs to be updated."""
            )

            software_spec_id = client.software_specifications.get_id_by_name(name)
            return software_spec_id

    try:
        base_software_spec_id = client.software_specifications.get_id_by_name(
            base_sofware_spec
        )

        meta_prop_sw_spec = {
            client.software_specifications.ConfigurationMetaNames.NAME: name,
            client.software_specifications.ConfigurationMetaNames.BASE_SOFTWARE_SPECIFICATION: {
                "guid": base_software_spec_id
            },
        }

        sw_spec_details = client.software_specifications.store(
            meta_props=meta_prop_sw_spec
        )
        software_spec_id = client.software_specifications.get_uid(sw_spec_details)

        for custom_package in custom_packages:
            if not is_zipfile(custom_package["file"]):
                raise MlflowException(
                    f"{custom_package['file']} is not a valid zip file."
                )
            meta_prop_pkg_extn = {
                client.package_extensions.ConfigurationMetaNames.NAME: custom_package[
                    "name"
                ],
                client.package_extensions.ConfigurationMetaNames.TYPE: "pip_zip",
            }

            pkg_extn_details = client.package_extensions.store(
                meta_props=meta_prop_pkg_extn, file_path=custom_package["file"]
            )

            pkg_extn_id = client.package_extensions.get_id(pkg_extn_details)

            client.software_specifications.add_package_extension(
                software_spec_id, pkg_extn_id
            )

    except Exception as e:
        LOGGER.exception(e)
        if software_spec_exists(client=client, sw_spec=name):
            delete_sw_spec(client, name)

        raise MlflowException(e)

    LOGGER.info(
        f"Successfully created {name} software specification with ID {software_spec_id}"
    )

    return software_spec_id



import logging
from typing import Dict, List, Optional

import mlflow
import pandas as pd
from ibm_watson_machine_learning.client import APIClient
from mlflow.deployments import BaseDeploymentClient
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import ENDPOINT_NOT_FOUND, INVALID_PARAMETER_VALUE

from mlflow_watsonml.config import Config
from mlflow_watsonml.utils import *
from mlflow_watsonml.wml import *

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

# CONSTANTS
DEFAULT_SOFTWARE_SPEC = "runtime-22.2-py3.10"
DEFAULT_MODEL_TYPE = "scikit-learn_1.1"


def target_help():
    # TODO: Improve
    help_string = (
        "The mlflow-watsonml plugin integrates IBM WatsonML "
        "with the MLFlow deployments API.\n\n"
        "Before using this plugin, you must set up a json file "
        "containing WML credentials and create an environment variable "
        "along with deployment space name. "
    )
    return help_string


def run_local(name, model_uri, flavor=None, config=None):
    # TODO: implement
    raise MlflowException("mlflow-watsonml does not currently support run_local.")


class WatsonMLDeploymentClient(BaseDeploymentClient):
    def __init__(self, target_uri: str = "watsonml", config: Optional[Dict] = None):
        """Initialize a WML `APIClient`. The method has an optional parameter called
        `config` which should have the WML credentials. If `config` is `None`, then
        the plugin will try to search for WML credentials in `.env` file or the
        environment variables.

        Parameters
        ----------
        target_uri : str, optional
            target uri for mlflow deployment, by default "watsonml"
        config : Optional[Dict], optional
            wml credentials, by default None
        """
        super().__init__(target_uri)

        self.wml_config = Config(config=config)
        self.connect(
            wml_credentials=self.wml_config["wml_credentials"],
            endpoint=self.wml_config["deployment_space_name"],
        )

    def connect(self, wml_credentials: Dict, endpoint: Optional[str] = None) -> None:
        """Connect to WML APIClient and set the default deployment space

        Parameters
        ----------
        wml_credentials : Dict
            WML Credentials
        endpoint : Optional[str]
            Deployment space name
        """
        try:
            client = APIClient(wml_credentials=wml_credentials)
            self.endpoint = endpoint
            LOGGER.info("Connected to WML Client successfully")

        except Exception as e:
            raise MlflowException(
                "Could not establish connection, check wml credentials." f"{e}",
                error_code=ENDPOINT_NOT_FOUND,
            )

        self._wml_client = client

    def get_wml_client(self, endpoint: Optional[str] = None) -> APIClient:
        """Returns WML API client

        Parameters
        ----------
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        APIClient
            WML client
        """
        client = self._wml_client

        if endpoint is not None:
            try:
                client = set_deployment_space(
                    client=client, deployment_space_name=endpoint
                )
            except Exception as e:
                raise MlflowException(e)

            return client

        elif self.endpoint is not None:
            try:
                client = set_deployment_space(
                    client=client, deployment_space_name=self.endpoint
                )
            except Exception as e:
                raise MlflowException(e)

            return client

        else:
            raise MlflowException(
                f"Deployment space name not set. Please set it using `endpoint` parameter"
            )

    def create_deployment(
        self,
        name: str,
        model_uri: str,
        flavor: Optional[str] = None,
        config: Optional[Dict] = None,
        endpoint: Optional[str] = None,
    ) -> Dict:
        """Deploy a model at `model_uri` to a WML target. this method blocks until
        deployment completes (i.e. until it's possible to perform inference with the deployment).
        In the case of conflicts (e.g. if it's not possible to create the specified deployment
        without due to conflict with an existing deployment), raises a
        :py:class:`mlflow.exceptions.MlflowException`.

        Parameters
        ----------
        name : str
            name of the deployment
        model_uri : str
            URI (local or remote) of the model
        flavor : Optional[str], optional
            flavor of the deployed model, by default None
        config : Optional[Dict], optional
            configuration parameters, by default None
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        Dict
            deployment details dictionary
        """
        if config is None:
            config = dict()

        client = self.get_wml_client(endpoint=endpoint)

        # check if a deployment by that name exists
        if deployment_exists(client=client, name=name) or model_exists(
            client=client, name=name
        ):
            raise MlflowException(
                f"Deplyment {name} already exists. Use `update_deployment()` or use a different name",
                error_code=INVALID_PARAMETER_VALUE,
            )

        model_object = mlflow.sklearn.load_model(model_uri=model_uri)

        software_spec_type = config.get("software_spec_type", DEFAULT_SOFTWARE_SPEC)
        software_spec_uid = client.software_specifications.get_id_by_name(
            software_spec_type
        )

        model_description = config.get("model_description", "no explanation")
        model_type = config.get("model_type", DEFAULT_MODEL_TYPE)

        model_details = store_model(
            client=client,
            model_object=model_object,
            software_spec_uid=software_spec_uid,
            name=name,
            model_description=model_description,
            model_type=model_type,
        )

        model_id = get_model_id_from_model_details(
            client=client, model_details=model_details
        )

        LOGGER.info("Stored Model Details = %s", model_details)
        LOGGER.info("Stored Model UID = %s", model_id)

        batch = config.get("batch", False)

        deployment_details = deploy_model(
            client=client,
            name=name,
            model_id=model_id,
            batch=batch,
        )

        return deployment_details

    def update_deployment(
        self,
        name: str,
        model_uri: Optional[str] = None,
        flavor: Optional[str] = None,
        config: Optional[Dict] = None,
        endpoint: Optional[str] = None,
    ) -> Dict:
        """Update the deployment with the specified name. You can update the URI of the model, the
        flavor of the deployed model (in which case the model URI must also be specified), and/or
        any WML-specific attributes of the deployment (via `config`). By default, this method
        blocks until deployment completes (i.e. until it's possible to perform inference
        with the updated deployment).

        Parameters
        ----------
        name : str
            Unique name of the deployment to update
        model_uri : Optional[str], optional
            URI of a new model to deploy, by default None
        flavor : Optional[str], optional
            new model flavor to use for deployment. If provided,
            ``model_uri`` must also be specified. If ``flavor`` is unspecified but
            ``model_uri`` is specified, a default flavor will be chosen and the
            deployment will be updated using that flavor., by default None
        config : Optional[Dict], optional
            dict containing updated WML-specific configuration for the
            deployment, by default None
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        Dict
            deployment details dictionary
        """
        self.delete_deployment(name=name, config=config, endpoint=endpoint)

        deployment_details = self.create_deployment(
            name=name,
            model_uri=model_uri,
            flavor=flavor,
            config=config,
            endpoint=endpoint,
        )

        return deployment_details

    def delete_deployment(
        self, name: str, config: Optional[Dict] = None, endpoint: Optional[str] = None
    ):
        """Delete the deployment with name ``name`` from WML. Deletion is idempotent
        (i.e. deletion does not fail if retried on a non-existent deployment).

        Parameters
        ----------
        name : str
            name of the deployment to delete
        config : Optional[Dict], optional
            configuration parameters, by default None
        endpoint : Optional[str], optional
            deployment space name, by default None
        """
        if config is None:
            config = dict()

        client = self.get_wml_client(endpoint=endpoint)

        if deployment_exists(client=client, name=name):
            delete_deployment(client=client, name=name)

        if model_exists(client=client, name=name):
            delete_model(client=client, name=name)

    def list_deployments(self, endpoint: Optional[str] = None):
        """List deployments. This method returns an unpaginated list of all deployments

        Parameters
        ----------
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        List[Dict]
            A list of dicts corresponding to deployments. Each dict is guaranteed to
            contain a 'name' key containing the deployment name. The other fields of
            the returned dictionary and their types follow WML deployment details convention.
        """
        return list_deployments(client=self.get_wml_client(endpoint=endpoint))

    def get_deployment(self, name: str, endpoint: Optional[str] = None):
        """Returns a dictionary describing the specified deployment, throwing a
        :py:class:`mlflow.exceptions.MlflowException` if no deployment exists with the provided
        name.
        The dict is guaranteed to contain a 'name' key containing the deployment name.
        The other fields of the returned dictionary and their types follow WML convention.

        Parameters
        ----------
        name : str
            name of the deployment to fetch
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        Dict
            A dict corresponding to the retrieved deployment. The dict is guaranteed to
            contain a 'name' key corresponding to the deployment name.
        """
        return get_deployment(client=self.get_wml_client(endpoint=endpoint), name=name)

    def predict(
        self,
        deployment_name: str,
        inputs: Optional[pd.DataFrame] = None,
        endpoint: Optional[str] = None,
    ):
        """Compute predictions on inputs using the specified deployment

        Parameters
        ----------
        deployment_name : str
            Name of deployment to predict against
        inputs : pd.DataFrame, optional
            Input data (or arguments) to pass to the deployment for inference,
            by default Optional[pd.DataFrame]
        endpoint : Optional[str], optional
            deployment space name, by default None

        Returns
        -------
        pd.DataFrame
            Model predictions as pandas.DataFrame
        """
        client = self.get_wml_client(endpoint=endpoint)

        scoring_payload = {
            self._wml_client.deployments.ScoringMetaNames.INPUT_DATA: [
                {"values": inputs}
            ]
        }

        deployment_id = get_deployment_id_from_deployment_name(
            client=client,
            deployment_name=deployment_name,
        )

        predictions = client.deployments.score(
            deployment_id=deployment_id, meta_props=scoring_payload
        )["predictions"]

        fields = predictions[0]["fields"]

        frames = []
        for prediction in predictions:
            frames.extend(prediction["values"])

        ans = pd.DataFrame(frames, columns=fields)

        return ans

    def explain(self, deployment_name=None, df=None, endpoint=None):
        raise NotImplementedError()

    def create_endpoint(self, name, config=None):
        raise NotImplementedError()

    def update_endpoint(self, endpoint, config=None):
        raise NotImplementedError()

    def delete_endpoint(self, endpoint):
        raise NotImplementedError()

    def list_endpoints(self):
        return list_endpoints(client=self.get_wml_client())

    def get_endpoint(self, endpoint):
        return get_endpoint(client=self.get_wml_client(), endpoint=endpoint)

    def create_custom_wml_spec(
        self,
        name: str,
        base_software_spec: str,
        custom_packages: List[Dict[str, str]],
        rewrite: bool = False,
        endpoint: Optional[str] = None,
    ):
        client = self.get_wml_client(endpoint=endpoint)
        software_spec_id = create_custom_software_spec(
            client=client,
            name=name,
            base_sofware_spec=base_software_spec,
            custom_packages=custom_packages,
            rewrite=rewrite,
        )

        return software_spec_id
